/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: configparser.c 153 2004-05-25 22:04:30Z pasi $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <glib.h>

#include "configparser.h"
#include "spamserv.h"
#include "structs.h"
#include "irc.h"

extern struct conf_t conf;


conf_entries conf_toks[] = {
	{"servername",     0,	STRING, NULL, (const void *)"stop.spam", 0, &conf.servername, sizeof(conf.servername)},
	{"remoteport",     0,	NUMBER, NULL, (const void *)7325, 0, &conf.remoteport, sizeof(conf.remoteport)},
	{"remoteserver",   1,	STRING, NULL, (const void *)NULL, 0, &conf.remoteserver, sizeof(conf.remoteserver)},
	{"remotepassword", 0,	STRING, NULL, (const void *)NULL, 0, &conf.remotepassword, sizeof(conf.remotepassword)},
	{"serverdesc",     0,	STRING, NULL, (const void *)"spamserv", 0, &conf.serverdesc, sizeof(conf.serverdesc)},
	{"reportchan",	   0,   STRING, NULL, (const void *)"#massads", 1, &conf.reportchan, sizeof(conf.reportchan)},
	{"infochan",	   0,	STRING, NULL, (const void *)"#massads", 1, &conf.infochan, sizeof(conf.infochan)},
	{"diepass",	   1,   STRING, NULL, (const void *)NULL, 1, &conf.diepass, sizeof(conf.diepass)},
	{"dbhost",	   1,	STRING, NULL, (const void *)NULL, 1, &conf.db.host, sizeof(conf.db.host)},
	{"dbuser",         1,   STRING, NULL, (const void *)NULL, 1, &conf.db.user, sizeof(conf.db.user)},
	{"dbpass",         1,   STRING, NULL, (const void *)NULL, 1, &conf.db.pass, sizeof(conf.db.pass)},
	{"dbname",         1,   STRING, NULL, (const void *)NULL, 1, &conf.db.database, sizeof(conf.db.database)},
	{"dbport",	   0,   NUMBER, NULL, (const void *)0, 1, &conf.db.port, sizeof(conf.db.port)},
	{"autokilllen",	   0,   NUMBER, NULL, (const void *)180, 1, &conf.akilllen, sizeof(conf.akilllen)},
	{"fork",	   0,   NUMBER, NULL, (const void *)1, 0, &conf.fork, sizeof(conf.fork)},
	{"logpath",	   1,	STRING, NULL, (const void *)NULL, 0, &conf.logpath, sizeof(conf.logpath)},
	{"loglevels",	   1,	STRING, NULL, (const void *)NULL, 0, &conf.loglevels, sizeof(conf.loglevels)},
	{(char *)NULL,     0,	NUMBER, NULL, (const void *)0, 0, 0, 0}
};

void  print_error(guint line, guint col, const char *msg);

void  print_error(guint line, guint col, const char *msg) {
	printf ("\nError in line %u, column %u: %s\n", line, col, msg);
}

void handle_hup(int signum __attr_unused__)
{
	i_send_globops("Got signal HUP! Rehashing!");
	s_rehash_config();
}

char *get_string_setting(const char *name)
{
	conf_entries *ce;
	static char *buf = NULL;

	if (buf != NULL) {
		free(buf);
		buf = NULL;
	}

	/* for s_cleanup, forces to not allocate anything */
	if (name == NULL)
		return NULL;
	
	for (ce = conf_toks; ce->name != NULL; ce++) {
		if ((strcasecmp(ce->name, name) == 0) &&
		    (ce->dt == STRING) && (ce->value != NULL)) {
			buf = g_strdup(ce->value);
			return buf;
		}
	}

	buf = malloc(1);
	buf[0] = '\0';
	return buf;
}

int get_int_setting(const char *name)
{
	conf_entries *ce;

	for (ce = conf_toks; ce->name != NULL; ce++) {
		if ((strcasecmp(ce->name, name)==0) &&
		    (ce->dt == NUMBER))
			return (int)ce->value;
	}
	return -1;
}

void read_config() {
	const char *filename = CONFIGFILE;
	int fd;
	GScanner *scanner;
	GTokenType ttype;
	char *name=NULL;
	conf_entries *ce;
	int *x;
	static int rehashonly=0;
	
	GScannerConfig config = {
		" \t\r\n",                     /* characters to skip */
		G_CSET_a_2_z "_/" G_CSET_A_2_Z, /* identifier start */
		G_CSET_a_2_z "_./" G_CSET_A_2_Z G_CSET_DIGITS,/* identifier cont. */
		"#\n",                         /* single line comment */
		FALSE,                         /* case_sensitive */
		TRUE,                          /* skip multi-line comments */
		TRUE,                          /* skip single line comments */
		FALSE,                         /* scan multi-line comments */
		TRUE,                          /* scan identifiers */
		TRUE,                          /* scan 1-char identifiers */
		FALSE,                         /* scan NULL identifiers */
		FALSE,                         /* scan symbols */
		FALSE,                         /* scan binary */
		FALSE,                         /* scan octal */
		TRUE,                          /* scan float */
		TRUE,                          /* scan hex */
		FALSE,                         /* scan hex dollar */
		TRUE,                          /* scan single quote strings */
		TRUE,                          /* scan double quite strings */
		TRUE,                          /* numbers to int */
		FALSE,                         /* int to float */
		TRUE,                          /* identifier to string */
		TRUE,                          /* char to token */
		FALSE,                         /* symbol to token */
		FALSE,                         /* scope 0 fallback */
		FALSE,			       /* store int64 */
		0			       /* padding_dummy by six,
						  makes gcc shut up. ;) */
	};
	

/* setdefault_settings */
	if (rehashonly == 0) {
		for (ce = conf_toks; ce->name != NULL; ce++) {
			if (ce->required == 1)
				ce->defvalue=ce->value = NULL;
			if (ce->dt == STRING) {
				if (ce->value != NULL)
					free(ce->value);
				ce->value = (ce->defvalue!=NULL) ? (void*)strdup(ce->defvalue) : NULL;
			} else ce->value = (void *)ce->defvalue;
		}
	}
 




	
	fd = open(filename, O_RDONLY);
	
	if (fd < 0) {
		s_fatal("Unable to open config file \"%s\".\n", filename);
		return;
	}
	
	scanner = g_scanner_new(&config);
	g_scanner_input_file(scanner, fd);
	scanner->input_name = filename;
	
	while ((ttype = g_scanner_get_next_token(scanner)) == G_TOKEN_STRING) {
		
		if (name != NULL)
			free(name);
		
		name = strdup(scanner->value.v_string);
		ttype = g_scanner_get_next_token (scanner);
		
		if (ttype != G_TOKEN_EQUAL_SIGN) {
			g_scanner_unexp_token(scanner,G_TOKEN_EQUAL_SIGN,
					NULL,NULL,NULL,NULL,FALSE);
			continue;
		}
	
		ttype = g_scanner_get_next_token(scanner);
		switch (ttype) {
			case G_TOKEN_INT:
				processint(name,scanner->value.v_int,rehashonly);
				break;
			case G_TOKEN_STRING:
				processchar(name,scanner->value.v_string,rehashonly);
				break;
			default:
				g_scanner_unexp_token (scanner,ttype,
						NULL,NULL,NULL,NULL,FALSE);
				break;
		}
	}
	if (name != NULL) {
		free(name);
		name=NULL;
	}
	
	g_scanner_destroy(scanner);
	close(fd);
	
	/* Ensure that all required variables have been found */
	for (ce = conf_toks; ce->name != NULL; ce++) {
		if ((ce->required == 1) && (ce->defvalue != (void*)1)) {
			s_fatal("Error: A required value has not "
					"been set. [%s]\n",ce->name);
		}
		
		if (ce->storeptr == NULL) continue;
		if ((ce->dt == STRING) && (ce->value != NULL) && (strlen(ce->value) < ce->maxstrlen)) {
                 	       strcpy(ce->storeptr, ce->value);
                } else {
			x=ce->storeptr;
			*x=(int)ce->value;
		}
	}
	if (rehashonly == 0) rehashonly++;
 }
 
 
 void processint(const char *name, int value, int rehash)
 {
	conf_entries *ce;
	
	for (ce = conf_toks; ce->name != NULL; ce++) {
		if (strcasecmp(ce->name,name) != 0)
			continue;
		
		if ((rehash == 1) && (ce->rehash != 1)) return;
		if (ce->dt != NUMBER) {
			printf("check config option [%s]. "
					"Expected to be a number.\n", name);
			return;
		}
		ce->value=(void *)value;
		if (ce->required == 1)
			ce->defvalue = (void*)1;
	}
 }
 
void processchar(const char *name, char *value, int rehash)
 {
 	conf_entries *ce;
	for (ce = conf_toks; ce->name != NULL; ce++) {
		if (strcasecmp(ce->name,name) != 0)
			continue;

		if ((rehash == 1) && (ce->rehash != 1)) return;
		if (ce->dt!=STRING) {
			printf("check config option [%s]. "
				"Expected to be a string\n", name);
			return;
		}
		if (ce->value != NULL)
			free(ce->value);
		
		ce->value = strdup(value);
		if (ce->required == 1)
			ce->defvalue=(void*)1;
	}
 }

void cleanup_config_strings(void)
{
	conf_entries *ce;
	for (ce = conf_toks; ce->name != NULL; ce++) {
		if ((ce->dt == STRING) && (ce->value != NULL)) {
			free(ce->value);
			ce->value = NULL;
		}
	}
}

