#include <glib.h>
#include <time.h>
#include <stdio.h>

#include "spamserv.h"
#include "structs.h"
#include "database.h"
#include "irc.h"
#include "log.h"
#include "util.h"
#include "pseudoclient.h"
#include "cycles.h"

#define MAXCYCLES	5

struct timeout_t {
	timestamp_t timestamp;
	TIMEOUT_FUNC tof;
	gpointer data; /* data to be passed to the function */
};

struct cycle_pseudo_t {
	irc_user user;
	/* next four reserved for compatibility with pclient
	 * struct */
	char **channels;
	unsigned int chanlistlen;
	unsigned int freechannels;
	int type;
	/* cycle count */
	int cycles;
};

typedef struct timeout_t timeout;
typedef struct cycle_pseudo_t cycle_bot;

GSList *cyclechans = NULL;
GSList *timeouts = NULL;

cycle_bot cycler;

static void to_check(gpointer *item, gpointer time);
static void to_clean(gpointer item, gpointer ignored __attr_unused__);
static void cycle_chan_act(gpointer data);
static void cycler_part_chan(char *chan);
static void cycler_send_join(char *chan);
static void cycler_mutate(void);
static void cycler_update_info(void);
static int cycle_chan_cmp(gpointer p1, gpointer p2);
static cycle_chan *find_cycle_chan(char *chan);

void initialize_cycle_checks(void)
{
	g_strlcpy(cycler.user.gcos, "IRCtoo Spam Check", MAXREALLEN);
	g_strlcpy(cycler.user.username, "ss", MAXUSERLEN);
	/* for /whois */
	cycler.chanlistlen = 0;
	s_add_cycle_pseudo((pclient *)&cycler);
	cycler_update_info();
	i_send_nick(&cycler.user);
}

static void cycler_part_chan(char *chan)
{
	i_send_part(&cycler.user, chan);
	cycler.cycles++;
	cycler_mutate();
}

static void cycler_send_join(char *chan)
{
	i_send_sjoin(&cycler.user, chan);
}

static void cycler_mutate(void)
{
	if (cycler.cycles < MAXCYCLES)
		return;
	i_send_quit(&cycler.user, "Disappearing");
	cycler_update_info();
	i_send_nick(&cycler.user);
}

/*
 * randomize a new host and nick for the bot */
static void cycler_update_info(void)
{
	char tmp[MAXNICKLEN];

	do {
		g_snprintf(tmp, MAXNICKLEN, "pseudo%d", 
				g_random_int_range(100, 999));
	} while (!s_is_nick_free(tmp));
	g_strlcpy(cycler.user.nick, tmp, MAXNICKLEN);
	g_strlcpy(cycler.user.host, s_print_random_bothost(), MAXHOSTLEN);
	cycler.cycles = 0;
}

/* actual stuff for cycle checks */

int cycle_chan_add(const char *chan, int delay, gboolean dbadd)
{
	cycle_chan *cc;

	g_return_val_if_fail(chan != NULL, -1);
	
	if (!isvalid_chan(chan))
		return -EINVALIDCHAN;
	
	if (dbadd == FALSE)
		if (database_add_cycle(chan, delay) < 0)
			return -EDBERROR;
	cc = g_malloc0(sizeof(cycle_chan));
	g_strlcpy(cc->channel, chan, MAXCHANLEN);
	if (delay == -1)
		cc->delay = DEFAULT_CYCLE * 60;
	else
		cc->delay = delay * 60; /* seconds */
	cyclechans = g_slist_prepend(cyclechans, cc);
	s_log(LOG_DEBUG, "Added cycle chan %s (%d)",
	      chan, delay);
	cycle_chan_schedule_check(cc);
	return 0;
}

int cycle_chan_del(char *chan)
{
	cycle_chan *cc;

	cc = find_cycle_chan(chan);
	if (cc == NULL)
		return -ENOTFOUND;
	if (database_del_cycle(chan) < 0)
		return -EDBERROR;
	cyclechans = g_slist_remove(cyclechans, cc);
	g_free(cc);
	return 0;
}

static void cycle_chan_act(gpointer data)
{
	cycle_chan *cc;
	
	g_return_if_fail(data != NULL);

	cc = data;
	if (cc->status == JOINING) {
		cycler_send_join(cc->channel);
		cc->status = PARTING;
		timeout_add(g_random_int_range(5, 30),
			    (TIMEOUT_FUNC)cycle_chan_act, cc);
	} else {
		cycler_part_chan(cc->channel);
		cycle_chan_schedule_check(cc);
	}
}

void cycle_list_to(irc_user *u, char *s)
{
	g_strdelimit(s, "*?", '%');
	database_list_cycles_to(u, s);
	i_send_notice(u, "End of CHANNEL List");
}

int cycle_chan_schedule_check(cycle_chan *cc)
{
	cc->status = JOINING;
	return timeout_add(cc->delay, (TIMEOUT_FUNC)cycle_chan_act, cc);
}

static int cycle_chan_cmp(gpointer p1, gpointer p2)
{
	return g_ascii_strncasecmp(((cycle_chan *)p1)->channel, (char *)p2,
				   MAXCHANLEN);
}

static cycle_chan *find_cycle_chan(char *chan)
{
	GSList *tmp;

	tmp = g_slist_find_custom(cyclechans, chan,
			          (GCompareFunc)cycle_chan_cmp);
	if (tmp == NULL)
		return NULL;
	else
		return tmp->data;
}

/* "library" for handling timeouts */

int timeout_add(unsigned long secs, TIMEOUT_FUNC tof, gpointer data)
{
	timeout *to;

	to = g_malloc0(sizeof(timeout));
	if (to == NULL)
		return -1;
	to->timestamp = time(NULL) + secs;
	to->tof = tof;
	to->data = data;
	timeouts = g_slist_prepend(timeouts, to);
	s_log(LOG_DEBUG, "Added timeout for %lu secs.", secs);
	return 0;
}

static void to_check(gpointer *item, gpointer time)
{
	timeout *to;
	time_t *t;

	to = *item;
	t = time;
	if (to->timestamp >= *t)
		return;
	(*to->tof)(to->data);
	/* we're done with this timeout, free and mark for
	 * deletation */
	g_free(to);
	*item = NULL;
}

void timeout_run(void)
{
	time_t now;

	now = time(NULL);
	s_gslist_foreach(timeouts, (GFunc)to_check, &now);	
	timeouts = g_slist_remove_all(timeouts, NULL);
}

static void to_clean(gpointer item, gpointer ignored __attr_unused__)
{
	g_free(item);
}

void timeout_clean(void)
{
	g_slist_foreach(timeouts, (GFunc)to_clean, NULL);
	g_slist_free(timeouts);
}
