/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: ui.c 153 2004-05-25 22:04:30Z pasi $ */

#include <stdlib.h>
#include <glib.h>
#include <string.h>

#include "spamserv.h"
#include "structs.h"
#include "server.h"
#include "irc.h"
#include "users.h"
#include "log.h"
#include "ui.h"
#include "magic.h"
#include "pseudoclient.h"
#include "util.h"
#include "configparser.h"
#include "formats.h"
#include "cycles.h"

#define MAXUICMDLEN	20

struct ui_cmd {
	char root[MAXUICMDLEN];
	char subcmd[MAXUICMDLEN];
	int switches;
	int depth;
	char *string;
	irc_user *by;
};

extern struct conf_t conf;

static int s_uicmd_parse(struct ui_cmd *, char **);
static int s_uicmd_parse_switch(int *, char *);
static char *s_uicmd_parse_crazy(char *, unsigned long *);
static char *s_uicmd_parse_cycleline(char *str, int *delay);
	
void s_uicmd_parse_cmdspam(struct ui_cmd *);
void s_uicmd_parse_cmdhelp(struct ui_cmd *);
void s_uicmd_parse_cmddie(struct ui_cmd *);
void s_uicmd_parse_cmdchannel(struct ui_cmd *);
void s_uicmd_parse_cmdrehash(struct ui_cmd *);
void s_uicmd_parse_cmdforbid(struct ui_cmd *);
void s_uicmd_parse_cmdcycle(struct ui_cmd *);
void s_uicmd_parse_cmddebug(struct ui_cmd *);

struct pccmd {
        char *cmd;
        void (*func) (struct ui_cmd *);
};

static struct pccmd cmdtabx[] = {
        {"SPAM", s_uicmd_parse_cmdspam},
        {"HELP", s_uicmd_parse_cmdhelp}, 
	{"DIE", s_uicmd_parse_cmddie},
	{"CHANNEL", s_uicmd_parse_cmdchannel},
	{"FORBID", s_uicmd_parse_cmdforbid},
	{"REHASH", s_uicmd_parse_cmdrehash},
	{"CYCLE", s_uicmd_parse_cmdcycle},
	{"DEBUG", s_uicmd_parse_cmddebug},
        {0, 0}
};

static int s_uicmd_parse_switch(int *switches, char *tok) {
	if (g_ascii_strcasecmp(tok, "autokill") == 0 ||
	    g_ascii_strcasecmp(tok, "akill") == 0) {
	   *switches |= SPAMSTR_AKILL;
	} else if (g_ascii_strcasecmp(tok, "kill") == 0) {
		*switches |= SPAMSTR_KILL;
	} else if (g_ascii_strcasecmp(tok, "report") == 0) {
		*switches |= SPAMSTR_REPORT;
	} else if (g_ascii_strcasecmp(tok, "channel") == 0 || 
		   g_ascii_strcasecmp(tok, "channels") == 0) {
		*switches |= SPAMSTR_CHANNEL;
	} else if (g_ascii_strcasecmp(tok, "matchinfo") == 0) {
		*switches |= SPAMSTR_SHOWINFO;
	} else {
		return -1;
	}
	return 1;
}

static int s_uicmd_parse(struct ui_cmd *scmd, char **toks)
{	
	int nexttok;
	
	memset(scmd->root, '\0', sizeof(scmd->root));
	memset(scmd->subcmd, '\0', sizeof(scmd->subcmd));
	nexttok = 0;
	scmd->string = NULL;
	if (toks[nexttok++] == NULL)
		return -1;
	g_strlcpy(scmd->root,
		 (toks[0][0] == ':') ? toks[0] + 1 : toks[0], MAXUICMDLEN);
	scmd->depth++;
	if (toks[nexttok] == NULL)
		return 0;
	if (toks[nexttok++][0] == '-') {
		return -1;
	}
	g_strlcpy(scmd->subcmd, toks[1], MAXUICMDLEN);
	scmd->depth++;
	/* don't move this under the following check, this is
	 * important for listing stuff.. */
	scmd->switches = 0;
	if (toks[nexttok] == NULL)
		return 0;
	while (toks[nexttok][0] == '-') {
		/* parse switch */
		if (s_uicmd_parse_switch(&scmd->switches,
		     toks[nexttok] + 1) == -1)
			return -2;
		nexttok++;
		if (toks[nexttok] == NULL)
			/* incorrect. */
			return -1;
		
	}
	/* we need this check because there might not be switches */
	if (toks[nexttok] == NULL)
		return -1;
	scmd->string = g_strjoinv(" ", toks + nexttok);
	return 0;
}

void s_parse_command(irc_msg *msg)
{
	struct ui_cmd scmd;
	struct pccmd *p_cmd;
	int err;

	scmd.by = &msg->user;
	/* +1 to skip over the recipient (SpamServ) */
	err = s_uicmd_parse(&scmd, msg->msgtoks + 1);
	if (err == -1) {
		i_send_notice(&msg->user, FORMAT_ERR,
			      s_print_err(-EINVALIDCMDFORMAT));
		return;
	} else if (err == -2) {
		i_send_notice(&msg->user, FORMAT_ERR,
				s_print_err(-EINVALIDPARAMS));
		return;
	}
	
	/* call the appropriate command handler */
	for (p_cmd = cmdtabx; p_cmd->cmd; p_cmd++) {
		 if (g_ascii_strcasecmp(scmd.root, p_cmd->cmd) == 0) {
			(void) (*p_cmd->func) (&scmd);
			 g_free(scmd.string);
			 return;
		}
	}
	i_send_notice(&msg->user, FORMAT_ERR, s_print_err(-EINVALIDCMD));
	g_free(scmd.string);		
}

/* ------------ put only s_uicmd_parse_cmd* below --------------
 * the ugly stuff you know. ;) */

void s_uicmd_parse_cmdhelp(struct ui_cmd *scmd)
{
	if (scmd->subcmd[0] == '\0') {
		(void)s_send_helpfile(scmd->by, "main");
		return;
	}
	(void)s_send_helpfile(scmd->by, scmd->subcmd);
}

void s_uicmd_parse_cmddebug(struct ui_cmd *scmd __attr_unused__)
{
#if 0
	GArray *lusers;
	irc_user *u;
	unsigned int i;
	
	/* the following code is fundementally broken, DO NOT USE. */
	if (!is_privileged_sa(scmd->by->nick))
		return;
	if (g_ascii_strcasecmp(scmd->subcmd, "search") == 0) {
		if (scmd->string == NULL)
			scmd->string = g_strdup("*");
		lusers = s_find_users_by_host(scmd->string);
		for (i = 0; i < lusers->len; i++) {
			u = &g_array_index(lusers, irc_user, i);
			u = s_fill_user_by_nick(u->nick, u);
			i_send_notice(scmd->by, "%d %s (%s@%s) ",
				      i, u->nick, u->username,
				      u->host);
		}
		g_array_free(lusers, FALSE);
	}
#endif 
}

void s_uicmd_parse_cmdrehash(struct ui_cmd *scmd)
{
	if (is_privileged_sa(scmd->by->nick)) {
		i_send_notice(scmd->by, "Rehashing..");
		s_rehash_config();
		i_send_globops("REHASH by %s", scmd->by->nick);
	} else {
		i_send_notice(scmd->by, FORMAT_ERR,
			      s_print_err(-EINSUFFPRIV_SA));
	}
}

void s_uicmd_parse_cmddie(struct ui_cmd *scmd)
{
	if (scmd->subcmd[0] == '\0' ||
	    strncmp(scmd->subcmd, conf.diepass, sizeof(conf.diepass)) != 0) {
		i_send_notice(scmd->by, "Incorrect DIE password.");
		return;
	}
	i_send_notice(scmd->by, "SpamServ shutting down.");
	i_send_globops("SpamServ shutting down, DIE by %s", scmd->by->nick);
	status.quitting = 1;
}

void s_uicmd_parse_cmdcycle(struct ui_cmd *scmd)
{
	int status, delay;
	char *s;
		
	if (scmd->subcmd[0] == '\0')
		i_send_notice(scmd->by, "Usage: SPAMSERV HELP CYCLE");
	else if (g_ascii_strcasecmp(scmd->subcmd, "add") == 0) {
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      "Missing channel name");
			return;
		}
		if ((s = s_uicmd_parse_cycleline(scmd->string,
				                 &delay)) == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      "Invalid delay time.");
			return;
		}
		if ((status = cycle_chan_add(s, delay, FALSE)) < 0)
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(status));
		else
			i_send_notice(scmd->by, "Cycle check added.");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "del") == 0) {
		if (scmd->string == NULL)
			i_send_notice(scmd->by, FORMAT_ERR,
				      "Missing channel name.");
		else {
			if ((status = cycle_chan_del(scmd->string)) < 0)
				i_send_notice(scmd->by, FORMAT_ERR,
					       s_print_err(status));
			else
				i_send_notice(scmd->by, "Removed.");
		}
	} else if (g_ascii_strcasecmp(scmd->subcmd, "list") == 0) {
		if (scmd->string == NULL)
			scmd->string = g_strdup("*");
		cycle_list_to(scmd->by, scmd->string);
	}
}

void s_uicmd_parse_cmdforbid(struct ui_cmd *scmd)
{
	int status;

	if (scmd->subcmd[0] == '\0') {
		i_send_notice(scmd->by, "Usage: SPAMSERV HELP FORBID");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "add") == 0) {
		if (scmd->string == NULL)
			scmd->string = g_strdup("No reason");
		if ((status = s_pseudochan_forbid(scmd->string,
			                         s_print_usermask(scmd->by,
						 TRUE))) < 0)
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(status));
		else
			i_send_notice(scmd->by, "Forbid added.");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "del") == 0) {
		if ((status = s_pseudochan_unforbid(scmd->string)) < 0)
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(status));
		else
			i_send_notice(scmd->by, "Forbid removed.");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "list") ==  0) {
		if (scmd->string == NULL)
			scmd->string = g_strdup("*");
		s_pseudochan_forbids_list_to(scmd->by, scmd->string);
	} else {
		i_send_notice(scmd->by, FORMAT_ERR, 
			      s_print_err(-EINVALIDCMD));
	}
}

void s_uicmd_parse_cmdchannel(struct ui_cmd *scmd)
{
	int status;
	
	if (scmd->subcmd[0] == '\0') {
		i_send_notice(scmd->by, "Usage: SPAMSERV HELP CHANNEL");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "add") == 0) {
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EMISSINGCHAN));
                        return;
		}
		if (!is_privileged_sa(scmd->by->nick)) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EINSUFFPRIV_SA));
		}
		status = s_invite_pseudoclient(scmd->string, "SpamServ", scmd->by);
		if (status < 0)
			i_send_notice(scmd->by,
				      FORMAT_ERR, s_print_err(status));
		else
			i_send_notice(scmd->by, "Channel added");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "del") == 0) {
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EMISSINGCHAN));
                        return;
		}
		if (!is_privileged_sa(scmd->by->nick)) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EINSUFFPRIV_SA));
		}
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EMISSINGCHAN));
		}
		if ((status = s_del_pseudochan(scmd->string)) < 0)
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(status));
		else
			i_send_notice(scmd->by, "Channel removed");

	} else if (g_ascii_strcasecmp(scmd->subcmd, "list") == 0) {
		if (scmd->string == NULL)
			scmd->string = g_strdup("*");
		s_list_pseudochans_to(scmd->by, scmd->string);
	} else if (g_ascii_strcasecmp(scmd->subcmd, "status") == 0) { 
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, FORMAT_ERR,
				      s_print_err(-EMISSINGCHAN));
		}
		i_send_notice(scmd->by, "Pseudoclient inhabiting '%s': %s",
				scmd->string, s_pseudo_at_chan(scmd->string));
	} else {
		i_send_notice(scmd->by, "Invalid command %s", scmd->subcmd);
	}
}

void s_uicmd_parse_cmdspam(struct ui_cmd *scmd)
{
	/* for parsing reason and comment */
	unsigned long id;
	char *s;

	int status; 
		
	if (scmd->subcmd[0] == '\0') {
		i_send_notice(scmd->by, "Usage: SPAMSERV HELP SPAM");
		return;
	}
	if (g_ascii_strcasecmp(scmd->subcmd, "add") == 0) {
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, "Incomplete command.");	
			return;
		}
		if (scmd->switches == 0)
			scmd->switches = SPAMSTR_DEFAULT;
		status = spam_str_add(scmd->string,
				      s_print_usermask(scmd->by, TRUE), 0,
			              scmd->switches, 0, 0, 0, FALSE, &id);
		if (status < 0)
			i_send_notice(scmd->by,
				      FORMAT_ERR, s_print_err(status));
		else if (status == 0)
			i_send_notice(scmd->by, "String added to database, ID: %lu",
				      id);
		else if (status == 1)
			i_send_notice(scmd->by, "Flags modified");

	} else if (g_ascii_strcasecmp(scmd->subcmd, "del") == 0) {
		status = spam_str_del_id(strtoul(scmd->string, NULL, 0));
		if (status < 0)
			i_send_notice(scmd->by,
				      FORMAT_ERR, s_print_err(status));
		else 
			i_send_notice(scmd->by, "Deleted.");

	} else if (g_ascii_strcasecmp(scmd->subcmd, "list") == 0) {
		/* by default, list all. note that this is really
		 * free'ed later. */
		if (scmd->string == NULL)
			scmd->string = g_strdup("*");
		spam_str_list_to(scmd->by, scmd->switches, scmd->string);
	} else if (g_ascii_strcasecmp(scmd->subcmd, "reason") == 0) {
		if (scmd->string == NULL) {
			i_send_notice(scmd->by, "Usage: SPAMSERV REASON <ID> <text>");
			return;
		}
		if ((s = s_uicmd_parse_crazy(scmd->string, &id)) == NULL) {
			i_send_notice(scmd->by, "Usage: SPAMSERV REASON <ID> <text>");
			return;
		}
		status = spam_str_add_reason_by_id(id, s);
		if (status < 0)
			i_send_notice(scmd->by, FORMAT_ERR, s_print_err(status));
		else
			i_send_notice(scmd->by, "Reason updated");
	} else if (g_ascii_strcasecmp(scmd->subcmd, "comment") == 0) {
                if (scmd->string == NULL) {
                        i_send_notice(scmd->by, "Usage: SPAMSERV COMMENT <ID> <text>");
                        return;
                }
                if ((s = s_uicmd_parse_crazy(scmd->string, &id)) == NULL) {
                        i_send_notice(scmd->by, "Usage: SPAMSERV COMMENT <ID> <text>");
                        return;
                }
                status = spam_str_add_comment_by_id(id, s);
                if (status < 0)
                        i_send_notice(scmd->by, FORMAT_ERR, s_print_err(status));
                else
                        i_send_notice(scmd->by, "Comment updated");
	} else {
		i_send_notice(scmd->by, "Invalid command %s", scmd->subcmd);
	}
}

/* TODO: added error checking for strtoul to functions below */

/* parses comment and reason strings (<id> <text>). Stores the 
 * id to integer pointed by int, and returns a pointer to the string.
 * Returns NULL on malformed command. */
static char *s_uicmd_parse_crazy(char *str, unsigned long *id)
{
	char *s;

	s = strchr(str, ' ');
	if (s == NULL || *(++s) == '\0')
		return NULL;
	*(s - 1) = '\0';
	/* we need this because strtoul() has the weirdest error
	 * handling */
	if (!g_ascii_isdigit(*str))
		return NULL;
	*id = strtoul(str, NULL, 0);
	return s;
}

static char *s_uicmd_parse_cycleline(char *str, int *delay)
{
	char *s;
	
	s = strchr(str, ' ');
	if (s == NULL || *(++s) == '\0') {
		/* default cycle time */
		*delay = -1;
		return str;
	}
	*(s - 1) = '\0';
	if (!g_ascii_isdigit(*s))
		return NULL;
	*delay = strtoul(s, NULL, 0);
	return str;
}
