/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: util.c 153 2004-05-25 22:04:30Z pasi $ */

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "spamserv.h"
#include "structs.h"
#include "users.h"
#include "pseudoclient.h"
#include "util.h"
#include "irc.h"

#define CODE_BOLD 0x02
#define CODE_UNDERLINE 0x1F
#define CODE_COLOUR 0x03
#define CODE_CR '\r'
#define CODE_LF '\n'

GSList *nicklist=NULL;

/* these (must) correspond to ones defined in spamserv.h.  */
static char *errors[] = 
{
	NULL,				  /* error 0 is not specified */
	"Database error",					/* 1  */
	"Database error, insufficient memory",			/* 2  */
	"Database connection failed",				/* 3  */
	"Not connected to database",				/* 4  */
	"Database Query Failed",				/* 5  */
	"Spam String already exists",				/* 6  */
	"Spam String specified is too wide, please compile a more " 
	"exact version",					/* 7  */
	"Invalid data",						/* 8  */
	"Error occurred",					/* 9  */
	"Not found",						/* 10 */
	"Invalid channel name",					/* 11 */
	"Missing channel name",					/* 12 */
	"Removal of channel failed",				/* 13 */
	"Insufficient privileges, must be SA or higher",	/* 14 */
	"Insufficient privileges, must be operator",		/* 15 */
	"Invalid command",					/* 16 */
	"Invalid command format",				/* 17 */
	"Connection closed",					/* 18 */
	"Read error",						/* 19 */
	"Object already exists in database.",			/* 20 */
	"Invalid command arguments.", 				/* 21 */
	NULL
};

void read_nicklist(void);
void s_foreach_free_generic(gpointer, gpointer);

char *strip_all(char *buffer)
{
	char *c,*c2, *c3;

	c=c2=buffer;
	do {
		if ((*c != CODE_BOLD) && (*c != CODE_UNDERLINE)
		    && (*c != CODE_COLOUR)  && (*c != CODE_CR)
		    && (*c != CODE_LF))
			*c2++=*c;
		else if (*c == CODE_COLOUR) {
			for (c++, c3=c+4;
			     (*c != '\0') && (c < c3) && (*c==','
			     || ((*c>='0')&&(*c<='9')));
			     c++);
			--c;
		}
	} while (*++c != '\0');
	*c2=0;
	return buffer;
}

char *s_print_usermask(irc_user *u, gboolean dofill)
{
	static char mask[512];
	
	if (dofill == TRUE)
		(void)s_fill_user_by_nick(u->nick, u);
	g_snprintf(mask, sizeof(mask), "%s!%s@%s",
		   u->nick, u->username, u->host);
	return mask;
}

char *s_print_random_bothost(void)
{
	static char hostmask[23];

	g_snprintf(hostmask, sizeof(hostmask), "pseudo-%d.stop.spam",
		   g_random_int_range(10000, 65535));
	return hostmask;
}

char *s_timestamp_to_human(time_t unixtime)
{
	static char hr[16];
	const struct tm *p_tm;
	
	if ((p_tm = gmtime(&unixtime)) == NULL)
		return "<N/A>";
	strftime(hr, 16, "%b %d %H:%M:%S", p_tm);
	return hr;
}

int isvalid_chan(const char *chan)
{
	if ((*chan != '#') && (*chan != '&')) return 0;
	if (strlen(chan) > 199) return 0;
	while (*++chan)
		if ((*chan == 7) || (*chan == ' ') || (*chan == ',')) return 0;
	return 1;
}

void read_nicklist()
{
	FILE *fp;
	char buffer[200], *c;
	
	fp=fopen("nicks.txt", "r");
	if (fp==NULL)
		return;
	srand(time(NULL));
	while (fgets(buffer, sizeof(buffer), fp)) {
		c=g_strdup(strip_all(buffer));
		/* FIXME: Use a better 'random placement' routine ?? */
		if (rand() % 2) {
			nicklist=g_slist_append(nicklist, c);
		} else {
			nicklist=g_slist_prepend(nicklist, c);
		}
	}
	fclose(fp);
	
}

void s_foreach_free_generic(gpointer data, gpointer ignored __attr_unused__)
{
	g_free(data);
}

void s_free_slist_generic(GSList **list)
{
	g_slist_foreach(*list, (GFunc)s_foreach_free_generic, NULL);
	g_slist_free(*list);
	*list = NULL;
}


/* 
 * XXX: I really don't know what is the _correct_ location
 * for this function. It really doesn't belong to neither of
 * users.c or pseudoclient.c because it deals with both. 
 * */
int s_is_nick_free(char *nick)
{
	if (s_find_user(nick) != NULL ||
			s_find_pseudoclient(nick) != NULL)
		return 0;
	return 1;
}

char *getnick()
{
	static char buffer[200];
	static unsigned long rdnl=0;

	if (rdnl==0) {
		read_nicklist();
		++rdnl;
	}
	nicklist=g_slist_next(nicklist);
	if (nicklist == NULL) {
		sprintf(buffer, "pseudo%lu", rdnl++);
	} else {
		strcpy(buffer, nicklist->data);
	}
	return buffer;
}

int s_send_helpfile(irc_user *to, const char *filex)
{
	FILE *helpfile;
	char *file;
	char buf[255];
	char filename[512];
	
	file = g_ascii_strdown(filex, -1);
	g_strcanon(file, "abcdefghijklmnopqrstuvwxyz", '_');
	g_snprintf(filename, sizeof(filename),
		   "%s/%s", HELPDIR, file);
	g_free(file);
	helpfile = fopen(filename, "r");
	if (helpfile == NULL) {
		i_send_notice(to, "No help for %s", filex);
		return -1;
	}
			
	while (fgets(buf, sizeof(buf), helpfile)) {
		if (ferror(helpfile)) {
			fclose(helpfile);
			i_send_notice(to, "Error while reading help file.");
			return -1;
		}
		i_send_notice(to, buf);
	}
	fclose(helpfile);
	i_send_notice(to, "End of Help");
	return 0;
}

char *s_print_err(int errnum)
{
	/* errors are always below zero (easier to check for)
	 * but we want them converted here */
	errnum = -errnum;
	if (errnum < 0 || errnum > MAXERR)
		return "Invalid errnum";
	return errors[errnum];
}

void s_gslist_foreach(GSList *list, GFunc func, gpointer data)
{
	GSList *tmp;
	
	for (; list; list = tmp) {
		tmp = list->next;
		(*func) (&(list->data), data);
	}
}
